# Generated by Django 5.2.11 on 2026-02-07 19:00

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='LegacyCalculationId',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('context_id', models.TextField()),
                ('calculation_record', models.TextField()),
                ('calculation_id', models.TextField()),
            ],
            options={
                'verbose_name': 'V1 Calculation ID',
                'verbose_name_plural': 'V1 Calculation IDs',
                'db_table': 'generic_app_calculationids',
                'managed': False,
            },
        ),
        migrations.CreateModel(
            name='LegacyCalculationLog',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('timestamp', models.DateTimeField()),
                ('trigger_name', models.TextField(blank=True, null=True)),
                ('message_type', models.TextField()),
                ('calculationId', models.TextField(db_column='calculationId')),
                ('message', models.TextField()),
                ('method', models.TextField()),
                ('is_notification', models.BooleanField()),
                ('calculation_record', models.TextField()),
            ],
            options={
                'verbose_name': 'V1 Calculation Log',
                'verbose_name_plural': 'V1 Calculation Logs',
                'db_table': 'generic_app_calculationlog',
                'ordering': ['-timestamp'],
                'managed': False,
            },
        ),
        migrations.CreateModel(
            name='LegacyUserChangeLog',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('user_name', models.TextField()),
                ('timestamp', models.DateTimeField()),
                ('message', models.TextField()),
                ('traceback', models.TextField(blank=True, null=True)),
                ('calculationId', models.TextField(db_column='calculationId')),
                ('calculation_record', models.TextField()),
            ],
            options={
                'verbose_name': 'V1 User Change Log',
                'verbose_name_plural': 'V1 User Change Logs',
                'db_table': 'generic_app_userchangelog',
                'ordering': ['-timestamp'],
                'managed': False,
            },
        ),
    ]
